#include "mini.h"

#define HeapStackSize       50000L
#define TrailCopySize       50000L

int     ErrorNumber;
FILE    *infile;
FILE    *outfile;
MW      infile_name;
MW      outfile_name;

void m_error(int errnr);

static CODEPTR firstclauseptr;

static int compile(MW *fname);

int main(int argc, char *argv[])
{
MW      fname;
size_t  heap_stack_size, trail_copy_size;

heap_stack_size = HeapStackSize;
trail_copy_size = TrailCopySize;
if (mm_initialise(heap_stack_size, trail_copy_size))
	{
	m_error(ErrorNumber);
	exit(1);
	}
firstclauseptr = 0;

infile = stdin;
outfile = stdout;
fname = infile_name = outfile_name = UserAtom;

find_atom("_mini",&fname);
(void) compile(&fname);

if (firstclauseptr != 0)
	m_error(mi_interpreter(firstclauseptr));
#ifdef Atari
(void) getchar();
#endif
return(Success);
}


static int compile(MW *fname)
{
MWPTR   copysave;
MW      term;
CODEPTR clauseptr;
int     err;

if (b_see(fname) != Success)
	return(ErrorNumber = FileOpenError);

copysave = copyptr;
for (;;)
	{
	copyptr = copysave;
	if ((b_eof() != Fail) || ((err = scan_term(&term)) == EndOfInputError))
		{
		(void) b_seen();
		copyptr = copysave;
		return(Success);
		}
	if (err == Success)
		{
		(void) compile_clause(term, FALSE, &clauseptr);
		if (firstclauseptr == 0)
			firstclauseptr = clauseptr;
		}
	}
}


void m_error(int errnr)
{
switch(errnr)
	{
	case Success:
		printf("No Error, Success\n");
		break;
	case Fail:
		printf("No Error, Fail\n");
		break;
	case SystemError:
		printf("System Error\n");
		break;
	case SyntaxError:
		printf("Syntax Error\n");
		break;
	case OperatorError:
		printf("Operator Precedence Error\n");
		break;
	case EndOfInputError:
		printf("Error: Unexpected End Of Input\n");
		break;
	case OutOfMemoryError:
		printf("Error: Out Of Memory\n");
		break;
	case FileOpenError:
		printf("Error Opening File\n");
		break;
	case UndefProcError:
		printf("Error: Calling undefined procedure\n");
		break;
	case NotImplementedError:
		printf("SystemError: Not yet implemented\n");
		break;
	case InvalidGoalError:
		printf("Error: Asserting invalid goal\n");
		break;
	case OutOfVarsError:
		printf("Error: Too many variables\n");
		break;
	case TooManyFilesError:
		printf("Error: Too many active files\n");
		break;
	case ExitTrueError:
		printf("Exit with success\n");
		break;
	case ExitFailError:
		printf("Exit with fail\n");
		break;
	case BuiltinInstError:
		printf("Wrong instantiated builtin argument\n");
		break;
	case BuiltinTypeError:
		printf("Wrong type of builtin argument\n");
		break;
	case BuiltinRangeError:
		printf("Builtin argument out of range\n");
		break;
	case BuiltinOpError:
		printf("Invalid operator definition\n");
		break;
	case StackOverflowError:
		printf("Overflow of environment stack\n");
		break;
	case CopyOverflowError:
		printf("Overflow of copy stack / trail stack\n");
		break;
	case NoBlockError:
		printf("Error: No matching block for exit_block\n");
		break;
	default:
		printf("Unknown Error\n");
	}
}
