/* mm.c     91/03/21   memory manager (initialisation, atoms, operators)      */
/*	  93/02/27   inclusion of assembler for the R3000		   */

#include "mini.h"
#ifdef mips
#include <mips/cachectl.h>
#endif

#define MinHeapCopySize     4096
#define MinTrailStackSize   4096
#define AtomTableSize       2048            /* must be 2^n */
#define FuncTableSize        512            /* must be 2^n */
#define OpTableSize           64            /* must be 2^n */
#define BipTableSize         128            /* must be 2^n > BuiltinSize */

#define XFX             0
#define XFY             1
#define YFX             2
#define FX              3
#define FY              4
#define XF              5
#define YF              6

#define LabelType       0
#define LoadType        1
#define ArithType       2
#define SubType         3
#define NorType         4
#define LoadImmType     5
#define ShiftType       6
#define MulDivType      7
#define MovHiLoType     8
#define BranchType      9
#define BranchZeroType 10
#define JumpType       11
#define JumpLinkType   12
#define MoveType       13
#define NopType        14
#define LogicalType    15

#define AsmTableSize    256
#define GlobalTableSize 128
#define RegTableSize     64

#define FunctSize           4                   /* size of functor block      *
#define FunctArity(f)       ((f)[0])             * is defined in mini.h       *
#define FunctIndex(f)       ((f)[1])             * is defined in mini.h       *
#define FunctProc(f)        (((MWPTR*)(f))[2])   * is defined in mini.h       */
#define FunctNext(f)        (((MWPTR*)(f))[3])  /* next with same hash value  */

#define LongCode            0,0
#define Header              LongCode, LongCode, LongCode, 0
#define Frame(s)            ((s)*SIZEMW)
#define FirstTemp(n)        h_fsttemp, ((n)*SIZEMW)
#define HeaderSize          5 * LopSize

#ifdef stackchange
#define InstrStackCheck     if(((MWPTR)instrptr)>=copyptr) \
                            return(ErrorNumber = OutOfMemoryError)
#else
#define InstrStackCheck     if(((MWPTR)instrptr)>=stackptr) \
                            return(ErrorNumber = OutOfMemoryError)
#endif

typedef struct OP {
	MW      op_next;		/* next operator with the same hash value */
	MW      op_name;		/* operator name  */
	int     op_assoc		/* operator associytivity */;
	int     op_prec;		/* operator precedence */
	} OP, *OPPTR;

typedef long    INSTR, *INSTRPTR;
typedef int     (*FUNC)(MW);

typedef struct ASMINSTR {
	MW      asm_atom_index;
	INSTR   asm_reg_code;
	INSTR   asm_imm_code;
	int     asm_instr;
} ASMINSTR;

typedef struct REGTABLE {
	MW      reg_atom;
	int     reg_nr;
} REGTABLE;

	
typedef struct GLOBALTABLE {
	MW      global_atom;
	MW      global_address;
} GLOBALTABLE;

	
MW      RuleAtom, CommaAtom, SColonAtom, VBarAtom, UserAtom, DotAtom, TrueAtom, CurlyBrackAtom;
MW      CutAtomIndex, NullAtomIndex, LCurlyAtomIndex, BlockAtomIndex, ExitBlockAtomIndex;
MW      PlusAtomIndex, MulAtomIndex, DivAtomIndex, ModAtomIndex;
MW      ShiftRightAtomIndex, ShiftLeftAtomIndex, BitAndAtomIndex, BitOrAtomIndex, BitNotAtomIndex, AbsAtomIndex;
MW      RuleAtomIndex, CommaAtomIndex, SColonAtomIndex, MinusAtomIndex;
MW      ForwardAtomIndex, BackwardAtomIndex, LabelAtomIndex, GlobalAtomIndex;
MWPTR   ForwardFunctor, BackwardFunctor, LabelFunctor, GlobalFunctor;
MW      UpperAtomIndex, UpperPlusAtomIndex, LowerAtomIndex;
MWPTR   UpperFunctor, UpperPlusFunctor, LowerFunctor;
MWPTR   DivFunctor, RuleFunctor, CommaFunctor, SColonFunctor, CurlyBrackFunctor;
MWPTR   CutFunctor, NullFunctor, BlockFunctor, ExitBlockFunctor;

static MW   xfxAtom, xfyAtom, yfxAtom, fxAtom, fyAtom, xfAtom, yfAtom;

static  MW      atom_table[AtomTableSize];
static  MWPTR   func_table[FuncTableSize];
static  MW      op_table[OpTableSize];
static  int     bip_table[BipTableSize];

static  MWPTR       labellist_start;
static  int	 asm_table_size, reg_table_size, global_table_size;
static  ASMINSTR    asm_table[AsmTableSize];
static  REGTABLE    reg_table[RegTableSize];
static  GLOBALTABLE global_table[GlobalTableSize];
static  MW	  CscellIndex;
static  MWPTR       CscellFunctor;

static  char	*reg_names[] = {"zz","at","v0","v1","a0","a1","a2","a3",
                                    "t0","t1","t2","t3","t4","t5","t6","t7",
                                    "s0","s1","s2","s3","s4","s5","s6","s7",
                                    "t8","t9","k0","k1","gp","sp","s8","ra"};

BIP_TAB builtin[BuiltinSize];

static int add_op(MW atom, int assoc, int prec);
static void insert_bip(char *nameptr, int arity, int bipnr, BIPPTR bipptr);
static void compile_bip(char *nameptr, int arity, CODEPTR codeptr);

int b_xtr_atom(MW arg[]);
int b_xtr_functor(MW arg[]);
int b_xtr_heapptr(MW arg[]);
int b_jump(MW arg[]);
int b_dump(MW arg[]);

extern int fail();
extern int unification();
extern int arg_unify();
extern int not_unify();
extern int int_compute();


static void init_assembler(void);
static void init_reg_table(char *stringptr, int nr);
static void init_global_table(char *stringptr, MW address);
static void init_asm_table(char *stringptr, INSTR rcode, INSTR icode, int type);
static int  find_reg_table(MW atom);
static int  find_instr(MW atom_index);

static CODE exit_block[] = {Header, Frame(0), LongCode, h_fsttmp0,
                            c_exit_block};


MWPTR       HeapStart, stackptr, StackStart, *TrailStart, CopyStart;

MWPTR       heapptr, copyptr, *trail;                   /* stackchange: no modification */
MWPTR       choicepntptr;

MWPTR		oldstackptr;

static b_break()
{
return(Success);
}


/* shows the avl-tree at rptr to level 'level' and depth 'depth' */
static void show_avl_tree(int level, int depth, MWPTR rptr)
{
int i, l;
MW m;

if (level == 0)
	{
	l = (1 << depth) - 1;
	for (i = 1; i < l; i++) putchar(' ');
	if (rptr != 0)
		{
		if (InstrRead((CODEPTR) rptr) == h_const)
			{
			m = LopRead(((CODEPTR) rptr) + 1);
			if (Int(m))
				printf("%2ld", (long) XtrInt(m));
			else
				printf("++");
			}
		else
			printf("**");
		if (AvlFlag(rptr) < 0)
			putchar('<');
		else if (AvlFlag(rptr) == 0)
			putchar('=');
		else
			putchar('>');
		}
	else
		printf("000");
	for (i = 0; i < l; i++) putchar(' ');
	return;
	}
if (rptr == 0)
	{
	show_avl_tree(level - 1, depth, 0);
	show_avl_tree(level - 1, depth, 0);
	return;
	}
show_avl_tree(level - 1, depth, Less(rptr));
show_avl_tree(level - 1, depth, Greater(rptr));
}


/* shows the avl-tree for arg[0] */
static b_show_avl_tree(MW arg[])
{
MWPTR   mptr;
MW      m, mi;
int     i;

m = arg[0];
ValDeref(m);
if (RefUndef(m))
	return(ErrorNumber = BuiltinInstError);
if (NotStruct(m))
	return(ErrorNumber = BuiltinTypeError);
mptr = XtrStruct(m);
if (XtrFunctor(mptr) != DivFunctor)
	return(ErrorNumber = BuiltinTypeError);
mptr = FirstArgPtr(mptr);
m = *mptr++;
ValDeref(m);
mi = *mptr;
ValDeref(mi);
if ((NotAtom(m)) || (NotInt(mi)))
	return(ErrorNumber = BuiltinTypeError);
if ((mi = XtrInt(mi)) < 0)
	return(ErrorNumber = BuiltinRangeError);
if ((mptr = find_func(XtrAtomIndex(m), mi)) == 0)
	return(ErrorNumber);
if ((mptr = find_proc(mptr)) == 0)
	return(ErrorNumber);
mptr = Less(mptr+1);
if (IsUndefProc(mptr))
	mptr = 0;
for (i = 0; i < 5; i++)
	{
	show_avl_tree(i, 5 - i, mptr);
	printf("\n");
	}
return(Success);
}


/* shows the current atom table */
static b_show_atom_table()
{
int     i, j, k;
MWPTR   aptr;

k = 0;
for (i = 0; i < AtomTableSize; i++)
	{
	j = 0;
	aptr = HeapStart + atom_table[i];
	while (aptr != HeapStart)
		{
		j++;
		if (j > 2)
			printf("%2d: %s\n", j, (char *) aptr);
		aptr = HeapStart + *(aptr-1);
		}
	k += j;
	if (j < 10)
		printf("%1d ", j);
	else
		printf("> ");
	if (((i + 1) & 63) == 0)
		printf("\n");
	}
printf("%d atoms\n", k);
return(Success);
}


/* initialises the memeory model: heap, copy, trail and environment stack */
/* heap and copy/environment stack use heap_copy_size*4 bytes;
   trail and environment/copy stack use trail_stack_size*4 bytes; 
	
   and initialises the bip- and atom-table */
int mm_initialise(size_t heap_copy_size, size_t trail_stack_size)
{
int i;
MW  op;

if (heap_copy_size < MinHeapCopySize)
	heap_copy_size = MinHeapCopySize;
if (trail_stack_size < MinTrailStackSize)
	trail_stack_size = MinTrailStackSize;
HeapStart = (MWPTR)malloc(sizeof(MW)*(heap_copy_size+trail_stack_size+6));
if (HeapStart == NULL)
	return(ErrorNumber = OutOfMemoryError);
HeapStart = (MWPTR)(((MW)HeapStart + 3) & ((MW) -4));
*HeapStart++ = 0;
*HeapStart++ = 0;
*HeapStart++ = 0;
*HeapStart++ = 0;
*HeapStart++ = 0;

#ifdef stackchange
	CopyStart   = HeapStart + heap_copy_size;
	StackStart  = CopyStart + trail_stack_size;
	TrailStart = (MWPTR*) CopyStart;
#else
	StackStart = HeapStart + heap_copy_size;  /* stackchange: heap_copy_size means heap_stack_size; name wasn't changed */
											  CopyStart  = StackStart + trail_stack_size; /* stackchange: trail_stack_size means trail_copy_size; name wasn't changed */
	TrailStart = (MWPTR*) StackStart;
#endif

#ifdef Ass
if (assinit() != 0)
	return(ErrorNumber = OutOfMemoryError);
#endif


stackptr = StackStart;

heapptr = HeapStart;
trail = TrailStart;
copyptr = CopyStart;

*((CODEPTR) HeapStart) = c_undefproc;
*(((CODEPTR) HeapStart) + 1) = h_undefproc;
heapptr++;

for (i = 0; i < AtomTableSize; i++)
	atom_table[i] = 0;
for (i = 0; i < FuncTableSize; i++)
	func_table[i] = 0;
for (i = 0; i < OpTableSize; i++)
	op_table[i] = 0;
for (i = 0; i < BipTableSize; i++)
	bip_table[i] = -1;
for (i = 0; i < BuiltinSize; i++)
	builtin[i].bt_arity = -1;

(void) find_atom(":-", &RuleAtom);
(void) find_atom(",", &CommaAtom);
(void) find_atom(";", &SColonAtom);
(void) find_atom("|", &VBarAtom);
(void) find_atom("user", &UserAtom);
(void) find_atom("true", &TrueAtom);
(void) find_atom("xfx", &xfxAtom);
(void) find_atom("xfy", &xfyAtom);
(void) find_atom("yfx", &yfxAtom);
(void) find_atom("fx",  &fxAtom);
(void) find_atom("fy ", &fyAtom);
(void) find_atom("xf",  &xfAtom);
(void) find_atom("yf",  &yfAtom);
(void) find_atom(".", &DotAtom);
(void) find_atom("{}", &CurlyBrackAtom);
(void) find_atom("{", &LCurlyAtomIndex);
(void) find_atom("!", &CutAtomIndex);
(void) find_atom("", &NullAtomIndex);
(void) find_atom("$block", &BlockAtomIndex);
(void) find_atom("$exit_block", &ExitBlockAtomIndex);
(void) find_atom("+", &PlusAtomIndex);
(void) find_atom("-", &MinusAtomIndex);
(void) find_atom("*", &MulAtomIndex);
(void) find_atom("/", &DivAtomIndex);
(void) find_atom("mod", &ModAtomIndex);
(void) find_atom("abs", &AbsAtomIndex);
(void) find_atom(">>", &ShiftRightAtomIndex);
(void) find_atom("<<", &ShiftLeftAtomIndex);
(void) find_atom("/\\", &BitAndAtomIndex);
(void) find_atom("\\/", &BitOrAtomIndex);
(void) find_atom("f", &ForwardAtomIndex);
(void) find_atom("b", &BackwardAtomIndex);
(void) find_atom("label", &LabelAtomIndex);
(void) find_atom("global", &GlobalAtomIndex);
(void) find_atom("upper", &UpperAtomIndex);
(void) find_atom("upper_plus", &UpperPlusAtomIndex);
(void) find_atom("lower", &LowerAtomIndex);

CurlyBrackFunctor = find_func(XtrAtomIndex(CurlyBrackAtom), (MW) 1);
LCurlyAtomIndex = XtrAtomIndex(LCurlyAtomIndex);
RuleAtomIndex = XtrAtomIndex(RuleAtom);
RuleFunctor = find_func(RuleAtomIndex, (MW) 2);
CommaAtomIndex = XtrAtomIndex(CommaAtom);
CommaFunctor = find_func(CommaAtomIndex, (MW) 2);
SColonAtomIndex = XtrAtomIndex(SColonAtom);
SColonFunctor = find_func(SColonAtomIndex, (MW) 2);
CutAtomIndex = XtrAtomIndex(CutAtomIndex);
CutFunctor = find_func(CutAtomIndex, (MW) 0);
NullAtomIndex = XtrAtomIndex(NullAtomIndex);
NullFunctor = find_func(NullAtomIndex, (MW) 0);
BlockAtomIndex = XtrAtomIndex(BlockAtomIndex);
BlockFunctor = find_func(BlockAtomIndex, (MW) 3);
ExitBlockAtomIndex = XtrAtomIndex(ExitBlockAtomIndex);
ExitBlockFunctor = find_func(ExitBlockAtomIndex, (MW) 1);
PlusAtomIndex = XtrAtomIndex(PlusAtomIndex);
MinusAtomIndex = XtrAtomIndex(MinusAtomIndex);
MulAtomIndex = XtrAtomIndex(MulAtomIndex);
DivAtomIndex = XtrAtomIndex(DivAtomIndex);
DivFunctor = find_func(DivAtomIndex, (MW) 2);
ModAtomIndex = XtrAtomIndex(ModAtomIndex);
AbsAtomIndex = XtrAtomIndex(AbsAtomIndex);
ShiftRightAtomIndex = XtrAtomIndex(ShiftRightAtomIndex);
ShiftLeftAtomIndex = XtrAtomIndex(ShiftLeftAtomIndex);
BitAndAtomIndex = XtrAtomIndex(BitAndAtomIndex);
BitOrAtomIndex = XtrAtomIndex(BitOrAtomIndex);
BitNotAtomIndex = XtrAtomIndex(BitNotAtomIndex);
BackwardAtomIndex = XtrAtomIndex(BackwardAtomIndex);
BackwardFunctor = find_func(BackwardAtomIndex, (MW) 1);
ForwardAtomIndex = XtrAtomIndex(ForwardAtomIndex);
ForwardFunctor = find_func(ForwardAtomIndex, (MW) 1);
LabelAtomIndex = XtrAtomIndex(LabelAtomIndex);
LabelFunctor = find_func(LabelAtomIndex, (MW) 1);
GlobalAtomIndex = XtrAtomIndex(GlobalAtomIndex);
GlobalFunctor = find_func(GlobalAtomIndex, (MW) 1);
UpperAtomIndex = XtrAtomIndex(UpperAtomIndex);
UpperFunctor = find_func(UpperAtomIndex, (MW) 1);
UpperPlusAtomIndex = XtrAtomIndex(UpperPlusAtomIndex);
UpperPlusFunctor = find_func(UpperPlusAtomIndex, (MW) 1);
LowerAtomIndex = XtrAtomIndex(LowerAtomIndex);
LowerFunctor = find_func(LowerAtomIndex, (MW) 1);

(void) add_op(RuleAtomIndex, XFX, 1200);
(void) add_op(RuleAtomIndex,  FX, 1200);
(void) find_atom("?-", &op);
(void) add_op(XtrAtomIndex(op), FX, 1200);
(void) add_op(SColonAtomIndex, XFY, 1100);
(void) add_op(CommaAtomIndex, XFY, 1000);
(void) find_atom("=", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("\\=", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("==", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("\\==", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("@<", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("@=<", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("@>", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("@>=", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("=..", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("is", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("=:=", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("=\\=", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("<", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("=<", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom(">", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom(">=", &op);
(void) add_op(XtrAtomIndex(op), XFX, 700);
(void) find_atom("+", &op);
(void) add_op(XtrAtomIndex(op), YFX, 500);
(void) find_atom("-", &op);
(void) add_op(XtrAtomIndex(op), YFX, 500);
(void) find_atom("/\\", &op);
(void) add_op(XtrAtomIndex(op), YFX, 500);
(void) find_atom("\\/", &op);
(void) add_op(XtrAtomIndex(op), YFX, 500);
(void) find_atom("*", &op);
(void) add_op(XtrAtomIndex(op), YFX, 400);
(void) find_atom("/", &op);
(void) add_op(XtrAtomIndex(op), YFX, 400);
(void) find_atom("<<", &op);
(void) add_op(XtrAtomIndex(op), YFX, 400);
(void) find_atom(">>", &op);
(void) add_op(XtrAtomIndex(op), YFX, 400);
(void) find_atom("mod", &op);
(void) add_op(XtrAtomIndex(op), YFX, 400);
(void) find_atom("pow", &op);
(void) add_op(XtrAtomIndex(op), XFX, 200);
(void) find_atom("-", &op);
(void) add_op(XtrAtomIndex(op), FY, 200);
(void) find_atom("\\", &op);
(void) add_op(XtrAtomIndex(op), FY, 200);
(void) find_atom("-->", &op);
(void) add_op(XtrAtomIndex(op), XFX, 1200);

(void) find_atom("exit_true", &op);
builtin[bip_exit_true].bt_name = XtrAtomIndex(op);
builtin[bip_exit_true].bt_arity = 0;
builtin[bip_exit_true].bt_bipptr = b_exit_true;

(void) find_atom("exit_fail", &op);
builtin[bip_exit_fail].bt_name = XtrAtomIndex(op);
builtin[bip_exit_fail].bt_arity = 0;
builtin[bip_exit_fail].bt_bipptr = b_exit_fail;

insert_bip("abolish",       1,  2, b_abolish);
insert_bip("arg",           3,  3, b_arg);
insert_bip("assert",        1,  4, b_assertz);
insert_bip("asserta",       1,  5, b_asserta);
insert_bip("assertz",       1,  6, b_assertz);
insert_bip("atom",          1,  7, b_atom);
insert_bip("atomic",        1,  8, b_atomic);
insert_bip("atom_length",   2,  9, b_atom_length);
insert_bip("atom_list",     2, 10, b_atom_list);
insert_bip("$block",        3, 11, b_block);
insert_bip("char_int",      2, 12, b_char_int);
insert_bip("compound",      1, 13, b_compound);
/* insert_bip("concat",        3, 13, b_concat); */
insert_bip("copy_term",     2, 14, b_copy_term);
insert_bip("cputime",       1, 15, b_cputime);
insert_bip("dec",           2, 16, b_dec);
insert_bip("display",       1, 17, b_display);
insert_bip("eof",           0, 18, b_eof);
/* insert_bip("$exit_block",   1, 19, b_exit_block); */
insert_bip("fail",          0, 20, b_fail);
insert_bip("functor",       3, 21, b_functor);
insert_bip("get",           1, 22, b_get);
insert_bip("get0",          1, 23, b_get0);
insert_bip("inc",           2, 24, b_inc);
insert_bip("integer",       1, 25, b_integer);
insert_bip("is",            2, 26, b_is);
insert_bip("list",          1, 27, b_list);
insert_bip("nil",           1, 28, b_nil);
insert_bip("nl",            0, 29, b_nl);
insert_bip("nonvar",        1, 30, b_nonvar);
insert_bip("number",        1, 31, b_number);
insert_bip("op",            3, 32, b_op);
insert_bip("put",           1, 33, b_put);
insert_bip("read",          1, 34, b_read);
insert_bip("see",           1, 35, b_see);
insert_bip("seeing",        1, 36, b_seeing);
insert_bip("seen",          0, 37, b_seen);
insert_bip("struct",        1, 38, b_struct);
insert_bip("tell",          1, 39, b_tell);
insert_bip("telling",       1, 40, b_telling);
insert_bip("told",          0, 41, b_told);
insert_bip("var",           1, 42, b_var);
insert_bip("write",         1, 43, b_write);
insert_bip("writeq",        1, 44, b_writeq);
insert_bip("=..",           2, 45, b_univ);
insert_bip("=",             2, 46, b_unify);
insert_bip("\\=",           2, 47, b_notunify);
insert_bip("=:=",           2, 48, b_int_eq);
insert_bip("=\\=",          2, 49, b_int_ne);
insert_bip("<",             2, 50, b_int_lt);
insert_bip("=<",            2, 51, b_int_le);
insert_bip(">",             2, 52, b_int_gt);
insert_bip(">=",            2, 53, b_int_ge);
insert_bip("==",            2, 54, b_equal);
insert_bip("\\==",          2, 55, b_notequal);
insert_bip("@<",            2, 56, b_less);
insert_bip("@>",            2, 57, b_greater);
insert_bip("@=<",           2, 58, b_equalless);
insert_bip("@>=",           2, 59, b_greaterequal);
insert_bip("$clause",       3, 60, b_sysclause);
insert_bip("$retract",      3, 61, b_sysretract);
insert_bip("$cur_pred",     2, 62, b_syscur_pred);
insert_bip("$xtr_atom",     2, 89, b_xtr_atom);
insert_bip("$xtr_functor",  2, 90, b_xtr_functor);
insert_bip("$dump",         2, 91, b_dump);
insert_bip("$jump",         2, 92, b_jump);
insert_bip("$assemble",     2, 93, b_assemble);
insert_bip("$xtr_heapptr",  1, 94, b_xtr_heapptr);
insert_bip("show_atom_table",0,95, b_show_atom_table);
insert_bip("show_avl_tree", 1, 96, b_show_avl_tree);
#ifdef Trace
insert_bip("inittrace",     0, 97, b_inittrace);
insert_bip("printtrace",    0, 98, b_printtrace);
#endif
insert_bip("break",         0, 99, b_break);

compile_bip("exit_block", 1, exit_block+HeaderSize);
init_assembler();
return(Success);
}

/* searches the atom stringptr in the stomtable atomptr */
int find_atom(char *stringptr, MWPTR atomptr)
{
int index;
char *cptr = stringptr;
MWPTR aptr;

index = *cptr++;
if (*cptr) {
	index = index ^ (*cptr++ << 7);
	if (*cptr) {
		index = index ^ (*cptr++ << 3);
	if (*cptr) {
		index = index ^ (*cptr++ << 1);
	if (*cptr) {
		index = index ^ (*cptr++ << 5);
	if (*cptr) {
		index = index ^ (*cptr++ << 2);
	}}}}}
aptr = HeapStart + atom_table[index = index & (AtomTableSize-1)];
while (aptr != HeapStart)
	{
	if (strcmp((char*)aptr, stringptr) == 0)
		{
		*atomptr = MakeAtom(aptr-HeapStart);
		return(Success);
		}
	aptr = HeapStart + *(aptr-1);
	}
*heapptr++ = atom_table[index];
*atomptr = MakeAtom(heapptr-HeapStart);
atom_table[index] = heapptr - HeapStart;
cptr = (char*)heapptr;

#ifdef stackchange
	while (((*cptr++ = *stringptr++) != 0) && (cptr < (char*) copyptr));
#else
	while (((*cptr++ = *stringptr++) != 0) && (cptr < (char*) stackptr));
#endif

heapptr += ((MWPTR)(cptr+SIZEMW-SIZECHAR) - heapptr);			/* stackchange: no modification */


																
#ifdef stackchange
	if (heapptr < copyptr)
		return(Success);
#else
	if (heapptr < stackptr)
		return(Success);
#endif

	
return(ErrorNumber = OutOfMemoryError);
}


/* searches the function with atomindex and arity in the function table */
MWPTR find_func(MW atom_index, MW arity)
{
MWPTR fptr;

fptr = func_table[atom_index & (FuncTableSize-1)];
while (fptr != 0)
	{
	if ((FunctIndex(fptr) == atom_index) && (FunctArity(fptr) == arity))
		return(fptr);
	fptr = FunctNext(fptr);
	}
fptr = heapptr;


#ifdef stackchange
	if ((heapptr += FunctSize) >= copyptr)
		{
		ErrorNumber = OutOfMemoryError;
		return((MWPTR) 0);
		}
#else
	if ((heapptr += FunctSize) >= stackptr)
		{
		ErrorNumber = OutOfMemoryError;
		return((MWPTR) 0);
		}
#endif

	
FunctIndex(fptr) = atom_index;
FunctArity(fptr) = arity;
FunctProc(fptr)  = (MWPTR) 0;
FunctNext(fptr)  = func_table[atom_index & (FuncTableSize-1)];
func_table[atom_index & (FuncTableSize-1)] = fptr;
return(fptr);
}


/* searches the procedure with functor 'functor' */
MWPTR find_proc(MWPTR functor)
{
MWPTR pptr;

if ((pptr = FunctProc(functor)) == 0)
	{
	pptr = heapptr += 4;

#ifdef stackchange
	if ((heapptr += ProcSize) >= copyptr)
		{
		ErrorNumber = OutOfMemoryError;
		return((MWPTR) 0);
		}
#else
	if ((heapptr += ProcSize) >= stackptr)
		{
		ErrorNumber = OutOfMemoryError;
		return((MWPTR) 0);
		}
#endif
		
	if (FunctArity(functor) == 0)
		ProcOffset(pptr) = HeapStart;
	else
		ProcOffset(pptr) = (MWPTR)((CODEPTR) HeapStart + 1);
	Equal(pptr + 1) = ProcOffset(pptr);
	Greater(pptr + 1) = ProcOffset(pptr);
	Less(pptr + 1) = ProcOffset(pptr);
	Indexing(pptr + 1) = x_general;
	ProcFunctor(pptr) = functor;
	FunctProc(functor) = pptr;
	}
return(pptr);
}


/* clause indexing algorithm */
void indexing(MWPTR rootptr)
{
Indexing(rootptr) = x_general;
if (IsUndefProc(Alternative(rootptr)))
	return;
if (Less(rootptr) != 0)
	return;
if (Greater(rootptr) == 0)
	return;
if (Equal(rootptr) == 0)
	return;
if (Equal(Greater(rootptr)) != 0)
	return;
if (Equal(Equal(rootptr)) != 0)
	return;
Indexing(rootptr) = x_nil_list;
}


/* inserts a clause in the current avl-tree */
static int avl_insert(MWPTR nextroot, MWPTR codep, MW codeval, int atbegin, MWPTR *prevp)
{
MW      findval;
MWPTR   findp, nextp, helpp;

if (*prevp == nextroot)
	{
	*prevp = codep;
	Equal(codep) = nextroot;
	Greater(codep) = nextroot;
	Less(codep) = nextroot;
	AvlFlag(codep) = 0;
	return(TRUE);
	}
findp = *prevp;
if (InstrRead((CODEPTR)findp) == h_const)
	findval = LopRead(((CODEPTR) findp) + 1);
else
	findval = (MW) FunctorFetch(((CODEPTR) findp) + 1);
if (codeval == findval)
	if (atbegin)
		{
		*prevp = codep;
		Equal(codep) = findp;
		Greater(codep) = Greater(findp);
		Less(codep) = Less(findp);
		AvlFlag(codep) = AvlFlag(findp);
		AvlFlag(findp) = 0;
		return(FALSE);
		}
	else
		{
		while (Equal(findp) != nextroot)
			findp = Equal(findp);
		Equal(findp) = codep;
		Equal(codep) = nextroot;
		AvlFlag(codep) = 0;
		return(FALSE);
		}
if (codeval < findval)
	if (avl_insert(nextroot, codep, codeval, atbegin, LessOffset(findp)))
		{
		switch (AvlFlag(findp))
			{
			case 1:
				AvlFlag(findp) = 0;
				return(FALSE);
			case 0:
				AvlFlag(findp) = -1;
				return(TRUE);
			default:
				nextp = Less(findp);
				if (AvlFlag(nextp) < 0)
					{
					Less(findp) = Greater(nextp);
					Greater(nextp) = findp;
					AvlFlag(findp) = 0;
					AvlFlag(nextp) = 0;
					*prevp = nextp;
					}
				else
					{
					helpp = Greater(nextp);
					Greater(nextp) = Less(helpp);
					Less(helpp) = nextp;
					Less(findp) = Greater(helpp);
					Greater(helpp) = findp;
					if (AvlFlag(helpp) < 0)
						AvlFlag(findp) = 1;
					else
						AvlFlag(findp) = 0;
					if (AvlFlag(helpp) > 0)
						AvlFlag(nextp) = -1;
					else
						AvlFlag(nextp) = 0;
					AvlFlag(helpp) = 0;
					*prevp = helpp;
					}
				return(FALSE);
			}
		}
	else
		return(FALSE);
if (avl_insert(nextroot, codep, codeval, atbegin, GreaterOffset(findp)))
	{
	switch (AvlFlag(findp))
		{
		case -1:
			AvlFlag(findp) = 0;
			return(FALSE);
		case 0:
			AvlFlag(findp) = 1;
			return(TRUE);
		default:
			nextp = Greater(findp);
			if (AvlFlag(nextp) > 0)
				{
				Greater(findp) = Less(nextp);
				Less(nextp) = findp;
				AvlFlag(findp) = 0;
				AvlFlag(nextp) = 0;
				*prevp = nextp;
				}
			else
				{
				helpp = Less(nextp);
				Less(nextp) = Greater(helpp);
				Greater(helpp) = nextp;
				Greater(findp) = Less(helpp);
				Less(helpp) = findp;
				if (AvlFlag(helpp) > 0)
					AvlFlag(findp) = -1;
				else
					AvlFlag(findp) = 0;
				if (AvlFlag(helpp) < 0)
					AvlFlag(nextp) = 1;
				else
					AvlFlag(nextp) = 0;
				AvlFlag(helpp) = 0;
				*prevp = helpp;
				}
			return(FALSE);
		}
	}
else
	return(FALSE);
}


/* inserts a clause in the current database */
void insert_clause(MWPTR procptr, CODEPTR clauseptr, int atbegin)
{
MWPTR   rootptr, rptr;

rootptr = procptr + 1;
rptr = ProcOffset(procptr);
if (IsUndefProc(rptr))
	{
	/* empty procedure */
	ProcOffset(procptr) = (MWPTR) clauseptr;
	Equal(rootptr) = 0;
	Greater(rootptr) = 0;
	Less(rootptr) = 0;
	}
else if (atbegin)
	{
	Alternative(clauseptr) = rptr;
	ProcOffset(procptr) = (MWPTR) clauseptr;
	if (ProcArity(procptr) == 0)
		return;
	/* search next root */
	rptr = Greater(rootptr); /* nil root */
	while ((rptr != 0) && (InstrRead((CODEPTR) rptr) == h_nil))
		rptr = Equal(rptr);
	switch(HeadMux(InstrRead(clauseptr)))
		{
		case HeadMux(h_const):
			(void) avl_insert(rptr, (MWPTR) clauseptr, LopRead(clauseptr + 1),
					  atbegin, LessOffset(rootptr));
			return;
		case HeadMux(h_nil):
			Equal(clauseptr) = Greater(rootptr);
			Greater(rootptr) = (MWPTR) clauseptr;
			return;
		case HeadMux(h_list):
			Equal(clauseptr) = Equal(rootptr);
			Equal(rootptr) = (MWPTR) clauseptr;
			return;
		case HeadMux(h_struct):
			(void) avl_insert(rptr, (MWPTR) clauseptr, (MW) FunctorFetch(clauseptr + 1),
					  atbegin, LessOffset(rootptr));
			return;
		case HeadMux(h_void):
		case HeadMux(h_fstvar):
		case HeadMux(h_fsttemp):
		case HeadMux(h_fsttmp0):
			Equal(clauseptr)   = Equal(rootptr);
			Greater(clauseptr) = Greater(rootptr);
			Less(clauseptr)    = Less(rootptr);
			Equal(rootptr)     = (MWPTR) clauseptr;
			Greater(rootptr)   = (MWPTR) clauseptr;
			Less(rootptr)      = (MWPTR) clauseptr;
			return;
		}
	}
else
	{
	/* search last root */
	for (;;)
		{
		rptr = Greater(rootptr);
		while ((rptr != 0) && (InstrRead((CODEPTR) rptr) == h_nil))
			rptr = Equal(rptr);
		if (rptr == 0)
			break;
		rootptr = rptr;
		}
	Alternative(ProcLastClause(procptr)) = (MWPTR) clauseptr;
	}
ProcLastClause(procptr) = (MWPTR) clauseptr;
Alternative(clauseptr) = 0;
if (ProcArity(procptr) == 0)
	return;
switch(HeadMux(InstrRead(clauseptr)))
	{
	case HeadMux(h_const):
		(void) avl_insert(0, (MWPTR) clauseptr, LopRead(clauseptr + 1),
				  atbegin, LessOffset(rootptr));
		return;
	case HeadMux(h_nil):
		if (Greater(rootptr) == 0)
			{
			Greater(rootptr) = (MWPTR) clauseptr;
			Equal(clauseptr) = 0;
			return;
			}
		rootptr = Greater(rootptr);
		/* fall through to case HeadMux(h_list) */
	case HeadMux(h_list):
		while (Equal(rootptr) != 0)
			rootptr = Equal(rootptr);
		Equal(rootptr) = (MWPTR) clauseptr;
		Equal(clauseptr) = 0;
		return;
	case HeadMux(h_struct):
		(void) avl_insert(0, (MWPTR) clauseptr, (MW) FunctorFetch(clauseptr + 1),
				  atbegin, LessOffset(rootptr));
		return;
	case HeadMux(h_void):
	case HeadMux(h_fstvar):
	case HeadMux(h_fsttemp):
	case HeadMux(h_fsttmp0):
		while ((CODEPTR) rootptr != clauseptr)
			{
			if (Equal(rootptr) == 0)
				Equal(rootptr) = (MWPTR) clauseptr;
			if (Greater(rootptr) == 0)
				Greater(rootptr) = (MWPTR) clauseptr;
			if (Less(rootptr) == 0)
				Less(rootptr) = (MWPTR) clauseptr;
			rootptr = Alternative(rootptr);
			}
		Equal(clauseptr) = 0;
		Greater(clauseptr) = 0;
		Less(clauseptr) = 0;
		return;
	}
}


/* true if prevp is less than the current node in the avl-tree */
static int balance_less(MWPTR *prevp)
{
MWPTR   p, p1, p2;
int     f;

p = *prevp;
switch (AvlFlag(p))
	{
	case -1:
		AvlFlag(p) = 0;
		return(TRUE);
	case 0:
		AvlFlag(p) = 1;
		return(FALSE);
	default:
		p1 = Greater(p);
		if ((f = AvlFlag(p1)) >= 0)
			{
			Greater(p) = Less(p1);
			Less(p1) = p;
			*prevp = p1;
			if (f == 0)
				{
				AvlFlag(p) = 1;
				AvlFlag(p1) = -1;
				return(FALSE);
				}
			AvlFlag(p) = 0;
			AvlFlag(p1) = 0;
			return(TRUE);
			}
		p2 = Less(p1);
		f = AvlFlag(p2);
		Less(p1) = Greater(p2);
		Greater(p2) = p1;
		Greater(p) = Less(p2);
		Less(p2) = p;
		if (f > 0)
			AvlFlag(p) = -1;
		else
			AvlFlag(p) = 0;
		if (f < 0)
			AvlFlag(p1) = 1;
		else
			AvlFlag(p1) = 0;
		AvlFlag(p2) = 0;
		*prevp = p2;
		return(TRUE);
	}
}


/* true if prevp is greater than the current node in the avl-tree */
static int balance_greater(MWPTR *prevp)
{
MWPTR   p, p1, p2;
int     f;

p = *prevp;
switch (AvlFlag(p))
	{
	case 1:
		AvlFlag(p) = 0;
		return(TRUE);
	case 0:
		AvlFlag(p) = -1;
		return(FALSE);
	default:
		p1 = Less(p);
		if ((f = AvlFlag(p1)) <= 0)
			{
			Less(p) = Greater(p1);
			Greater(p1) = p;
			*prevp = p1;
			if (f == 0)
				{
				AvlFlag(p) = -1;
				AvlFlag(p1) = 1;
				return(FALSE);
				}
			AvlFlag(p) = 0;
			AvlFlag(p1) = 0;
			return(TRUE);
			}
		p2 = Greater(p1);
		f = AvlFlag(p2);
		Greater(p1) = Less(p2);
		Less(p2) = p1;
		Less(p) = Greater(p2);
		Greater(p2) = p;
		if (f < 0)
			AvlFlag(p) = 1;
		else
			AvlFlag(p) = 0;
		if (f > 0)
			AvlFlag(p1) = -1;
		else
			AvlFlag(p1) = 0;
		AvlFlag(p2) = 0;
		*prevp = p2;
		return(TRUE);
	}
}


/* deletes a node from the current avl-tree */
static int delete_node(MWPTR nextroot, MWPTR *prevr)
{
MWPTR   r;

if ((r = *prevr) == nextroot)
	return(balance_greater(prevr));
if (delete_node(nextroot, GreaterOffset(r)))
	return(balance_greater(prevr));
return(FALSE);
}


/* deletes the current avl-tree */
static int avl_delete(MWPTR nextroot, MWPTR codep, MW codeval, MWPTR *prevp)
{
MW      findval;
MWPTR   p, q;

p = *prevp;
if (InstrRead((CODEPTR)p) == h_const)
	findval = LopRead(((CODEPTR) p) + 1);
else
	findval = (MW) FunctorFetch(((CODEPTR) p) + 1);
if (codeval < findval)
	if (avl_delete(nextroot, codep, codeval, LessOffset(p)))
		return(balance_less(prevp));
	else
		return(FALSE);
if (codeval > findval)
	if (avl_delete(nextroot, codep, codeval, GreaterOffset(p)))
		return(balance_greater(prevp));
	else
		return(FALSE);
if (codep != p)
	{
	while (Equal(p) != codep)
		p = Equal(p);
	Equal(p) = Equal(codep);
	Equal(codep) = 0;
	return(FALSE);
	}
if (Equal(codep) != nextroot)
	{
	p = Equal(codep);
	*prevp = p;
	AvlFlag(p) = AvlFlag(codep);
	Greater(p) = Greater(codep);
	Less(p) = Less(codep);
	Equal(codep) = 0;
	return(FALSE);
	}
Equal(codep) = 0;
if (Greater(codep) == nextroot)
	{
	*prevp = Less(codep);
	return(TRUE);
	}
if (Less(codep) == nextroot)
	{
	*prevp = Greater(codep);
	return(TRUE);
	}
q = p = Less(codep);
while (Greater(p) != nextroot)
	p = Greater(q = p);
*prevp = p;
AvlFlag(p) = AvlFlag(codep);
Greater(p) = Greater(codep);
if (q != p)
	{
	Greater(q) = Less(p);
	Less(p) = Less(codep);
	if (delete_node(q, LessOffset(p)))
		return(balance_less(prevp));
	return(FALSE);
	}
return(balance_less(prevp));
}


/* deletes a clause from the current database */
static void delete_clause(MWPTR procptr, CODEPTR clauseptr)
{
MWPTR   rootptr, cptr, rptr, pptr;

pptr = rootptr = procptr + 1;
cptr = ProcOffset(procptr);
if (IsUndefProc(cptr))
	return; /* can only retract once */
Indexing(pptr) = x_general;
if (ProcArity(procptr) != 0)
	{
	rptr = Equal(clauseptr);
	while ((rptr != 0) &&
	       (InstrRead((CODEPTR) rptr) != h_fsttmp0) &&
	       (InstrRead((CODEPTR) rptr) != h_fsttemp) &&
	       (InstrRead((CODEPTR) rptr) != h_fstvar) &&
	       (InstrRead((CODEPTR) rptr) != h_void))
		rptr = Equal(rptr);
	}
if ((CODEPTR) cptr == clauseptr)
	{
	if (Alternative(clauseptr) == 0)
		{
		ProcLastClause(procptr) = 0;
		if (ProcArity(procptr) == 0)
			{
			ProcOffset(procptr) = HeapStart;
			return;
			}
		ProcOffset(procptr) = (MWPTR)(((CODEPTR)HeapStart) + 1);
		Equal(rootptr) = ProcOffset(procptr);
		Greater(rootptr) = ProcOffset(procptr);
		Less(rootptr) = ProcOffset(procptr);
		return;
		}
	ProcOffset(procptr) = Alternative(clauseptr);
	Alternative(clauseptr) = 0;
	if (ProcArity(procptr) == 0)
		return;
	}
else
	{
	while ((pptr != 0) && ((CODEPTR) Alternative(pptr) != clauseptr))
		pptr = Alternative(pptr);
	if (pptr == 0)
		return; /* can only retract once */
	if (Alternative(clauseptr) == 0)
		ProcLastClause(procptr) = pptr;
	Alternative(pptr) = Alternative(clauseptr);
	Alternative(clauseptr) = 0;
	if (ProcArity(procptr) == 0)
		return;
	/* search root */
	for (;;)
		{
		cptr = Greater(rootptr);
		while ((cptr != rptr) && (InstrRead((CODEPTR) cptr) == h_nil))
			cptr = Equal(cptr);
		if ((cptr == rptr) || ((CODEPTR) cptr == clauseptr))
			break;
		rootptr = rptr;
		}
	}
switch(HeadMux(InstrRead(clauseptr)))
	{
	case HeadMux(h_const):
		(void) avl_delete(rptr, (MWPTR) clauseptr, LopRead(clauseptr + 1),
				  LessOffset(rootptr));
		return;
	case HeadMux(h_nil):
		if ((CODEPTR) Greater(rootptr) == clauseptr)
			{
			Greater(rootptr) = Equal(clauseptr);
			Equal(clauseptr) = 0;
			return;
			}
		rootptr = Greater(rootptr);
		/* fall through to case HeadMux(h_list) */
	case HeadMux(h_list):
		while ((CODEPTR) Equal(rootptr) != clauseptr)
			rootptr = Equal(rootptr);
		Equal(rootptr) = Equal(clauseptr);
		Equal(clauseptr) = 0;
		return;
	case HeadMux(h_struct):
		(void) avl_delete(rptr, (MWPTR) clauseptr, (MW) FunctorFetch(clauseptr + 1),
				  LessOffset(rootptr));
		return;
	case HeadMux(h_void):
	case HeadMux(h_fstvar):
	case HeadMux(h_fsttmp0):
	case HeadMux(h_fsttemp):
		Equal(clauseptr) = 0;
		Greater(clauseptr) = 0;
		Less(clauseptr) = 0;
		cptr = rootptr;
		for (;;)
			{
			if (Equal(cptr) == (MWPTR) clauseptr)
				Equal(cptr) = rptr;
			if (Greater(cptr) == (MWPTR) clauseptr)
				Greater(cptr) = rptr;
			if (Less(cptr) == (MWPTR) clauseptr)
				Less(cptr) = rptr;
			if (cptr == pptr)
				break;
			cptr = Alternative(cptr);
			}
		while (Alternative(cptr) != 0)
			{
			cptr = Alternative(cptr);
			switch(HeadMux(InstrRead(((CODEPTR)cptr))))
				{
				case HeadMux(h_const):
					(void) avl_insert(rptr, (MWPTR) cptr, LopRead(((CODEPTR) cptr) + 1),
							  FALSE, LessOffset(rootptr));
					continue;
				case HeadMux(h_nil):
					Equal(cptr) = rptr;
					if (Greater(rootptr) == rptr)
						{
						Greater(rootptr) = cptr;
						continue;
						}
					pptr = Greater(rootptr);
					while (Equal(pptr) != rptr)
						pptr = Equal(pptr);
					Equal(pptr) = cptr;
					continue;
				case HeadMux(h_list):
					Equal(cptr) = rptr;
					pptr = rootptr;
					while (Equal(pptr) != rptr)
						pptr = Equal(pptr);
					Equal(pptr) = cptr;
					continue;
				case HeadMux(h_struct):
					(void) avl_insert(rptr, (MWPTR) cptr, (MW) FunctorFetch(((CODEPTR) cptr) + 1),
							  FALSE, LessOffset(rootptr));
					continue;
				default:
					break;
				}
			break;
			}
		return;
	}
}


static int add_op(MW atomindex, int assoc, int prec)
{
OPPTR optr;
int   offset;

if ((assoc == FX) || (assoc == FY))
	offset = 1;
else
	offset = 0;
optr = (OPPTR) (HeapStart + op_table[(atomindex+offset)&(OpTableSize-1)]);
while (optr != (OPPTR) HeapStart)
	{
	if (optr->op_name==atomindex)
		return(ErrorNumber = BuiltinOpError);
	optr = (OPPTR) (HeapStart + optr->op_next);
	}
optr = (OPPTR) heapptr;
heapptr += ((sizeof(OP)+SIZEMW-SIZECHAR)/SIZEMW);

#ifdef stackchange
	if (heapptr < copyptr)
		{
		optr->op_next = op_table[(atomindex+offset)&(OpTableSize-1)];
		op_table[(atomindex+offset)&(OpTableSize-1)] = ((MWPTR)optr) - HeapStart;
		optr->op_name = atomindex;
		optr->op_assoc = assoc;
		optr->op_prec = prec;
		return(Success);
		}
#else
	if (heapptr < stackptr)
		{
		optr->op_next = op_table[(atomindex+offset)&(OpTableSize-1)];
		op_table[(atomindex+offset)&(OpTableSize-1)] = ((MWPTR)optr) - HeapStart;
		optr->op_name = atomindex;
		optr->op_assoc = assoc;
		optr->op_prec = prec;
		return(Success);
		}
#endif

return(ErrorNumber = OutOfMemoryError);
}



int is_prefix_op(MW atomindex, int *oprec, int *rprec)
{
OPPTR optr;

optr = (OPPTR) (HeapStart + op_table[(atomindex+1)&(OpTableSize-1)]);
while (optr != (OPPTR) HeapStart)
	{
	if ((optr->op_name == atomindex) &&
	    ((optr->op_assoc == FX) || (optr->op_assoc == FY)))
		{
		*oprec = optr->op_prec;
		if (optr->op_assoc == FX)
			*rprec = *oprec - 1;
		else
			*rprec = *oprec;
		return(TRUE);
		}
	optr = (OPPTR) (HeapStart + optr->op_next);
	}
return(FALSE);
}


int is_postfix_op(MW atomindex, int *lprec, int *oprec)
{
OPPTR optr;

optr = (OPPTR) (HeapStart + op_table[(atomindex)&(OpTableSize-1)]);
while (optr != (OPPTR) HeapStart)
	{
	if ((optr->op_name == atomindex) && (optr->op_assoc >= XF))
		{
		*oprec = optr->op_prec;
		if (optr->op_assoc == XF)
			*lprec = *oprec - 1;
		else
			*lprec = *oprec;
		return(TRUE);
		}
	optr = (OPPTR) (HeapStart + optr->op_next);
	}
return(FALSE);
}


int is_infix_op(MW atomindex, int *lprec, int *oprec, int *rprec)
{
OPPTR optr;

optr = (OPPTR) (HeapStart + op_table[(atomindex)&(OpTableSize-1)]);
while (optr != (OPPTR) HeapStart)
	{
	if ((optr->op_name == atomindex) && (optr->op_assoc <= YFX))
		{
		*oprec = optr->op_prec;
		*lprec = *oprec - 1;
		*rprec = *oprec - 1;
		if (optr->op_assoc == XFY)
			*rprec = *oprec;
		else if (optr->op_assoc == YFX)
			*lprec = *oprec;
		return(TRUE);
		}
	optr = (OPPTR) (HeapStart + optr->op_next);
	}
return(FALSE);
}


/* inserts the builtin-predicate bipnr (nameptr and arity) in the builtin-table */
static void insert_bip(char *nameptr, int arity, int bipnr, BIPPTR bipptr)
{
MW      name;
int i;

if ((bipnr < 0) || (bipnr >= BuiltinSize) || (builtin[bipnr].bt_arity >= 0))
	{
	m_error(SystemError);
	exit(1);
	}
(void) find_atom(nameptr, &name);
name = XtrAtomIndex(name);
builtin[bipnr].bt_name = name;
builtin[bipnr].bt_arity = arity;
builtin[bipnr].bt_bipptr = bipptr;
i = (int) ((name + arity) & (BipTableSize-1));
while (bip_table[i] >= 0)
	i = (i + 11) & (BipTableSize-1);
bip_table[i] = bipnr;
}


/* true, if the builtin predicate I (index and arity) is found in the builtin-table */
int find_bip(MW index, MW arity)
{
int i, bipnr;

i = (int) ((index + arity) & (BipTableSize-1));
while ((bipnr = bip_table[i]) >= 0)
	{
	if ((builtin[bipnr].bt_name == index) && 
	    (builtin[bipnr].bt_arity == arity))
		return(bipnr);
	i = (i + 11) & (BipTableSize-1);
	}
return(bipnr);
}


int b_op(MW arg[])
{
MW  op;
MW  m;
int assoc, prior;

m = arg[0];
ValDeref(m);
if (NotInt(m))
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
prior = (int) XtrInt(m);
if ((prior < 0) || (prior > MaxPriority))
	return(ErrorNumber = BuiltinRangeError);
m = arg[1];
ValDeref(m);
if (NotAtom(m))
	if (RefUndef(m))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
if (m == xfxAtom)
	assoc = XFX;
else if (m == xfyAtom)
	assoc = XFY;
else if (m == yfxAtom)
	assoc = YFX;
else if (m == fxAtom)
	assoc = FX;
else if (m == fyAtom)
	assoc = FY;
else if (m == xfAtom)
	assoc = XF;
else if (m == yfAtom)
	assoc = YF;
else
	return(ErrorNumber = BuiltinRangeError);
op = arg[2];
ValDeref(op);
if (NotAtom(op))
	if (RefUndef(op))
		return(ErrorNumber = BuiltinInstError);
	else
		return(ErrorNumber = BuiltinTypeError);
if (op == CommaAtom)
	return(ErrorNumber = BuiltinRangeError);
op = XtrAtomIndex(op);
return(add_op(op, assoc, prior));
}


/* deletes the procedure proc from the database */
static int abolish(MW proc)
{
MWPTR   mptr;
MW      m, mi;

if (RefUndef(proc))
	return(ErrorNumber = BuiltinInstError);
if (NotStruct(proc))
	return(ErrorNumber = BuiltinTypeError);
mptr = XtrStruct(proc);
if (XtrFunctor(mptr) != DivFunctor)
	return(ErrorNumber = BuiltinTypeError);
mptr = FirstArgPtr(mptr);
m = *mptr++;
ValDeref(m);
mi = *mptr;
ValDeref(mi);
if ((NotAtom(m)) || (NotInt(mi)))
	return(ErrorNumber = BuiltinTypeError);
if ((mi = XtrInt(mi)) < 0)
	return(ErrorNumber = BuiltinRangeError);
if ((mptr = find_func(XtrAtomIndex(m), mi)) == 0)
	return(ErrorNumber);
if ((mptr = find_proc(mptr)) == 0)
	return(ErrorNumber);
if (mi == 0)
	ProcOffset(mptr) = HeapStart;
else
	ProcOffset(mptr) = (MWPTR) ((CODEPTR) HeapStart + 1);
Equal(mptr+1) = ProcOffset(mptr);
Greater(mptr+1) = ProcOffset(mptr);
Less(mptr+1) = ProcOffset(mptr);
return(Success);
}


/* deletes the procedure arg[0] from the database */
int b_abolish(MW arg[])
{
MWPTR   mptr;
MW      m;

m = arg[0];
ValDeref(m);
while (List(m))
	{
	mptr = XtrList(m);
	m = *mptr++;
	ValDeref(m);
	if (abolish(m) != Success)
		return(ErrorNumber);
	m = *mptr;
	ValDeref(m);
	}
return(abolish(m));
}


static int clause_retract(MW arg[], int retract)
{
MWPTR   mptr, procptr, functor;
MW      head, m, clause;
CODEPTR cptr;

head = arg[0];
ValDeref(head);
if (Atom(head))
	{
	if ((functor = find_func(XtrAtomIndex(head), (MW) 0)) == 0)
		return(ErrorNumber);
	}
else if (Struct(head))
	functor = XtrFunctor(XtrStruct(head));
else if (RefUndef(head))
	return(ErrorNumber = BuiltinInstError);
else
	return(ErrorNumber = BuiltinTypeError);
procptr = FunctProc(functor);
m = *(mptr = arg + 2);
StructDeref(m, mptr);
if (RefUndef(m))
	{
	if (procptr == 0)
		cptr = 0;
	else
		{
		cptr = (CODEPTR) ProcOffset(procptr);
		if (IsUndefProc(cptr))
			cptr = 0;
		}
	}
else
	cptr = (CODEPTR) XtrList(m);
if (cptr == 0)
	{
	DerefChoicePntPtr;
	return(Fail);
	}
*mptr = MakeList(Alternative(cptr));
clause_decompile(functor, cptr, &clause);
if (Struct(clause) && (XtrFunctor(mptr=XtrStruct(clause)) == RuleFunctor))
	{
	if (unify(arg, FirstArgPtr(mptr)) != Success)
		return(Fail);
	if (unify(arg+1, FirstArgPtr(mptr) + 1) != Success)
		return(Fail);
	}
else
	{
	if (unify(arg, &clause) != Success)
		return(Fail);
	if (unify(arg+1, &TrueAtom) != Success)
		return(Fail);
	}
if (retract)
	delete_clause(procptr, cptr);
return(Success);
}


int b_sysclause(MW arg[])
{
return(clause_retract(arg, FALSE));
}


int b_sysretract(MW arg[])
{
return(clause_retract(arg, TRUE));
}


int b_syscur_pred(MW arg[])
{
MWPTR   mptr, hptr, aptr, functor;
MW      m, head;
MW      arity;
int     index;

head = *(hptr = arg);
StructDeref(head, hptr);
if (RefUndef(head))
	{
	Trail(hptr);
	CreateStruct(hptr, DivFunctor, 2);
	aptr = hptr + 1;
	head = MakeUndef(hptr);
	arity = MakeUndef(aptr);
	}
else if (Struct(head))
	{
	hptr = XtrStruct(head);
	if (XtrFunctor(hptr) != DivFunctor)
		return(ErrorNumber = BuiltinTypeError);
	head = *(hptr = FirstArgPtr(hptr));
	arity = *(aptr = hptr + 1);
	StructDeref(head, hptr);
	StructDeref(arity, aptr);
	if ((NotRefUndef(head) && NotAtom(head)) ||
	    (NotRefUndef(arity) && NotInt(arity)))
		return(ErrorNumber = BuiltinTypeError);
	}
else
	return(ErrorNumber = BuiltinTypeError);
m = *(mptr = arg + 1);
StructDeref(m, mptr);
if (RefUndef(head))
	{
	if (RefUndef(m))
		functor = func_table[index = 0];
	else
		{
		functor = XtrList(m);
		index = (int) (FunctIndex(functor) & (FuncTableSize-1));
		functor = FunctNext(functor);
		}
	for (;;)
		{
		while ((functor == 0) && (++index < FuncTableSize))
			functor = func_table[index];
		if (index >= FuncTableSize)
			break;
		if (FunctProc(functor) == 0)
			{
			functor = FunctNext(functor);
			continue;
			}
		if (RefUndef(arity))
			{
			Trail(aptr);
			*aptr = MakeInt(FunctArity(functor));
			}
		else if (XtrInt(arity) != FunctArity(functor))
			{
			functor = FunctNext(functor);
			continue;
			}
		Trail(hptr);
		*hptr = MakeAtom(FunctIndex(functor));
		*mptr = MakeList(functor);
		return(Success);
		}
	}
else
	{
	if (RefUndef(m))
		functor = func_table[XtrAtomIndex(head) & (FuncTableSize-1)];
	else
		functor = FunctNext(XtrList(m));
	while (functor != 0)
		{
		*mptr = MakeList(functor);
		if (FunctProc(functor) != 0)
			if (RefUndef(arity))
				{
				Trail(aptr);
				*aptr = MakeInt(FunctArity(functor));
				return(Success);
				}
			else if (XtrInt(arity) == FunctArity(functor))
				return(Success);
		functor = FunctNext(functor);
		}
	}
DerefChoicePntPtr;
return(Fail);
}


/* compiles the builtin predicate (nameptr and arity) to indermediate code at codeptr */
static void compile_bip(char *nameptr, int arity, CODEPTR codeptr)
{
int     result;
MW      name;
MWPTR   pptr;

if ((result = find_atom(nameptr, &name)) != Success)
	{
	m_error(result);
	exit(1);
	}
if ((pptr = find_func(XtrAtomIndex(name), (MW) arity)) == 0)
	{
	m_error(ErrorNumber);
	exit(1);
	}
if ((pptr = find_proc(pptr)) == 0)
	{
	m_error(ErrorNumber);
	exit(1);
	}
insert_clause(pptr, codeptr, FALSE);
}


/* initialises the inline-assembler */
static void init_assembler(void)
{
(void) find_atom("cscell", &CscellIndex);
CscellIndex = XtrAtomIndex(CscellIndex);
CscellFunctor = find_func(CscellIndex, (MW) 2);

reg_table_size = asm_table_size = global_table_size = 0;

init_reg_table("zero", 0);
init_reg_table("at",   1);
init_reg_table("v0",   2);
init_reg_table("v1",   3);
init_reg_table("a0",   4);
init_reg_table("a1",   5);
init_reg_table("a2",   6);
init_reg_table("a3",   7);
init_reg_table("t0",   8);
init_reg_table("t1",   9);
init_reg_table("t2",  10);
init_reg_table("t3",  11);
init_reg_table("t4",  12);
init_reg_table("t5",  13);
init_reg_table("t6",  14);
init_reg_table("t7",  15);
init_reg_table("s0",  16);
init_reg_table("s1",  17);
init_reg_table("s2",  18);
init_reg_table("s3",  19);
init_reg_table("s4",  20);
init_reg_table("s5",  21);
init_reg_table("s6",  22);
init_reg_table("s7",  23);
init_reg_table("t8",  24);
init_reg_table("t9",  25);
init_reg_table("k0",  26);
init_reg_table("k1",  27);
init_reg_table("gp",  28);
init_reg_table("sp",  29);
init_reg_table("s8",  30);
init_reg_table("ra",  31);
init_reg_table("nullreg", 0);
init_reg_table("at",   1);
init_reg_table("returnvalue",   2);
init_reg_table("structstart4",   3);
init_reg_table("argframeptr",   4);
init_reg_table("structstart6",   5);
init_reg_table("headvarreg6",   6);
init_reg_table("headvarreg5",   7);
init_reg_table("copystptr",   8);
init_reg_table("value2",   9);
init_reg_table("value1",  10);
init_reg_table("structstart0",  11);
init_reg_table("structstart1",  12);
init_reg_table("structstart2",  13);
init_reg_table("structstart3",  14);
init_reg_table("choicepointptr",  15);
init_reg_table("headframeptr",  16);
init_reg_table("headvarreg1",  17);
init_reg_table("listtagreg",  18);
init_reg_table("reftagreg",  19);
init_reg_table("headvarreg3",  20);
init_reg_table("headvarreg4",  21);
init_reg_table("headvarreg0",  22);
init_reg_table("goalframeptr",  23);
init_reg_table("temporary",  24);
init_reg_table("trailptr",  25);
init_reg_table("structstart8",  28);
init_reg_table("stackptr",  29);	    /*stackchange: no modification*/
init_reg_table("headvarreg2",  30);
init_reg_table("headvarreg7",  5);

init_global_table("abolish",       (MW) b_abolish);
init_global_table("arg",	   (MW) b_arg);
init_global_table("assert",	(MW) b_assertz);
init_global_table("asserta",       (MW) b_asserta);
init_global_table("assertz",       (MW) b_assertz);
init_global_table("atom",	  (MW) b_atom);
init_global_table("atomic",	(MW) b_atomic);
init_global_table("atom_length",   (MW) b_atom_length);
init_global_table("atom_list",     (MW) b_atom_list);
init_global_table("break",	 (MW) b_break);
init_global_table("char_int",      (MW) b_char_int);
init_global_table("compound",      (MW) b_compound);
init_global_table("copy_term",     (MW) b_copy_term);
init_global_table("cputime",       (MW) b_cputime);
init_global_table("dec",	   (MW) b_dec);
init_global_table("display",       (MW) b_display);
init_global_table("eof",	   (MW) b_eof);
init_global_table("functor",       (MW) b_functor);
init_global_table("get",	   (MW) b_get);
init_global_table("get0",	  (MW) b_get0);
init_global_table("inc",	   (MW) b_inc);
init_global_table("integer",       (MW) b_integer);
init_global_table("is",	    (MW) b_is);
init_global_table("list",	  (MW) b_list);
init_global_table("nil",	   (MW) b_nil);
init_global_table("nl",	    (MW) b_nl);
init_global_table("nonvar",	(MW) b_nonvar);
init_global_table("number",	(MW) b_number);
init_global_table("op",	    (MW) b_op);
init_global_table("put",	   (MW) b_put);
init_global_table("read",	  (MW) b_read);
init_global_table("see",	   (MW) b_see);
init_global_table("seeing",	(MW) b_seeing);
init_global_table("seen",	  (MW) b_seen);
init_global_table("struct",	(MW) b_struct);
init_global_table("tell",	  (MW) b_tell);
init_global_table("telling",       (MW) b_telling);
init_global_table("told",	  (MW) b_told);
init_global_table("var",	   (MW) b_var);
init_global_table("write",	 (MW) b_write);
init_global_table("writeq",	(MW) b_writeq);
init_global_table("=..",	   (MW) b_univ);
init_global_table("=",	     (MW) b_unify);
init_global_table("\\=",	   (MW) b_notunify);
init_global_table("=:=",	   (MW) b_int_eq);
init_global_table("=\\=",	  (MW) b_int_ne);
init_global_table("<",	     (MW) b_int_lt);
init_global_table("=<",	    (MW) b_int_le);
init_global_table(">",	     (MW) b_int_gt);
init_global_table(">=",	    (MW) b_int_ge);
init_global_table("==",	    (MW) b_equal);
init_global_table("\\==",	  (MW) b_notequal);
init_global_table("@<",	    (MW) b_less);
init_global_table("@>",	    (MW) b_greater);
init_global_table("@=<",	   (MW) b_equalless);
init_global_table("@>=",	   (MW) b_greaterequal);
init_global_table("fail",	  (MW) fail);
init_global_table("unification",   (MW) unification);
init_global_table("arg_unify",     (MW) arg_unify);
init_global_table("not_unify",     (MW) not_unify);
init_global_table("int_compute",   (MW) int_compute);
init_global_table("copyptr",       (MW) &copyptr);		  /* stackchange: no modification */
init_global_table("trail",	 (MW) &trail);
init_global_table("choicepntptr",  (MW) &choicepntptr);
init_global_table("oldstackptr",   (MW) &oldstackptr);
init_global_table("stackptr",      (MW) &stackptr);

init_asm_table("label",  (INSTR) 0,  (INSTR) 0x00000000L, LabelType);
init_asm_table("lw",     (INSTR) 0,  (INSTR) 0x8c000000L, LoadType);
init_asm_table("sw",     (INSTR) 0,  (INSTR) 0xac000000L, LoadType);
init_asm_table("add",    (INSTR) 32, (INSTR) 0x20000000L, ArithType);
init_asm_table("addu",   (INSTR) 33, (INSTR) 0x24000000L, ArithType);
init_asm_table("sub",    (INSTR) 34, (INSTR) 0x20000000L, SubType);
init_asm_table("subu",   (INSTR) 35, (INSTR) 0x24000000L, SubType);
init_asm_table("and",    (INSTR) 36, (INSTR) 0x30000000L, LogicalType);
init_asm_table("or",     (INSTR) 37, (INSTR) 0x34000000L, LogicalType);
init_asm_table("xor",    (INSTR) 38, (INSTR) 0x38000000L, LogicalType);
init_asm_table("nor",    (INSTR) 39, (INSTR) 0x00000000L, NorType);
init_asm_table("slt",    (INSTR) 42, (INSTR) 0x28000000L, ArithType);
init_asm_table("sltu",   (INSTR) 43, (INSTR) 0x2c000000L, ArithType);
init_asm_table("lui",    (INSTR) 37, (INSTR) 0x3c000000L, LoadImmType);
init_asm_table("sll",    (INSTR)  4, (INSTR) 0x00000000L, ShiftType);
init_asm_table("srl",    (INSTR)  6, (INSTR) 0x00000002L, ShiftType);
init_asm_table("sra",    (INSTR)  7, (INSTR) 0x00000003L, ShiftType);
init_asm_table("mult",   (INSTR) 24, (INSTR) 0x00000000L, MulDivType);
init_asm_table("multu",  (INSTR) 25, (INSTR) 0x00000000L, MulDivType);
init_asm_table("div",    (INSTR) 26, (INSTR) 0x00000000L, MulDivType);
init_asm_table("divu",   (INSTR) 27, (INSTR) 0x00000000L, MulDivType);
init_asm_table("mfhi",   (INSTR) 16, (INSTR) 0x00000000L, MovHiLoType);
init_asm_table("mthi",   (INSTR) 17, (INSTR) 0x00000000L, MovHiLoType);
init_asm_table("mflo",   (INSTR) 18, (INSTR) 0x00000000L, MovHiLoType);
init_asm_table("mtlo",   (INSTR) 19, (INSTR) 0x00000000L, MovHiLoType);
init_asm_table("beq",    (INSTR)  0, (INSTR) 0x10000000L, BranchType);
init_asm_table("bne",    (INSTR)  0, (INSTR) 0x14000000L, BranchType);
init_asm_table("blez",   (INSTR)  0, (INSTR) 0x18000000L, BranchZeroType);
init_asm_table("bgtz",   (INSTR)  0, (INSTR) 0x1c000000L, BranchZeroType);
init_asm_table("bltz",   (INSTR)  0, (INSTR) 0x04000000L, BranchZeroType);
init_asm_table("bgez",   (INSTR)  0, (INSTR) 0x04010000L, BranchZeroType);
init_asm_table("bltzal", (INSTR)  0, (INSTR) 0x04100000L, BranchZeroType);
init_asm_table("bgezal", (INSTR)  0, (INSTR) 0x04110000L, BranchZeroType);
init_asm_table("j",      (INSTR)  8, (INSTR) 0x08000000L, JumpType);
init_asm_table("jal",    (INSTR)  9, (INSTR) 0x0c000000L, JumpLinkType);
init_asm_table("move",   (INSTR) 37, (INSTR) 0x34000000L, MoveType);
init_asm_table("nop",    (INSTR) 37, (INSTR) 0x34000000L, NopType);
}


static void init_reg_table(char *stringptr, int nr)
{
MW  atom;
int i;

(void) find_atom(stringptr, &atom);
i = reg_table_size;
while ((--i >= 0) && (atom < reg_table[i].reg_atom))
	reg_table[i+1] = reg_table[i];
reg_table[i+1].reg_atom = atom;
reg_table[i+1].reg_nr = nr;
reg_table_size++;
}




static void init_global_table(char *stringptr, MW address)
{
MW  atom;
int i;

(void) find_atom(stringptr, &atom);
i = global_table_size;
while ((--i >= 0) && (atom < global_table[i].global_atom))
	global_table[i+1] = global_table[i];
global_table[i+1].global_atom = atom;
global_table[i+1].global_address = address;
global_table_size++;
}


static void init_asm_table(char *stringptr, INSTR rcode, INSTR icode, int type)
{
MW  atom, atom_index;
int i;

(void) find_atom(stringptr, &atom);
atom_index = XtrAtomIndex(atom);
i = asm_table_size;
while ((--i >= 0) && (atom_index < asm_table[i].asm_atom_index))
	asm_table[i+1] = asm_table[i];
asm_table[i+1].asm_atom_index = atom_index;
asm_table[i+1].asm_reg_code = rcode;
asm_table[i+1].asm_imm_code = icode;
asm_table[i+1].asm_instr = type;
asm_table_size++;
}


static int find_reg(MW atom)
{
int low, high, i;

low = 0;
high = reg_table_size;
while (low < high)
	{
	i = (high + low) >> 1;
	if (atom == reg_table[i].reg_atom)
		return i;
	if (atom < reg_table[i].reg_atom)
		high = i;
	else
		low = i + 1;
	}
return -1;
}


static int find_instr(MW atom_index)
{
int low, high, i;

low = 0;
high = asm_table_size;
while (low < high)
	{
	i = (high + low) >> 1;
	if (atom_index == asm_table[i].asm_atom_index)
		return i;
	if (atom_index < asm_table[i].asm_atom_index)
		high = i;
	else
		low = i + 1;
	}
return -1;
}


static MW find_global_label(MW atom)
{
int low, high, i;

ValDeref(atom);
if (NotAtom(atom))
	return 0;
low = 0;
high = global_table_size;
while (low < high)
	{
	i = (high + low) >> 1;
	if (atom == global_table[i].global_atom)
		return global_table[i].global_address;
	if (atom < global_table[i].global_atom)
		high = i;
	else
		low = i + 1;
	}
return 0;
}


static INSTRPTR find_label(MWPTR label)
{
MWPTR labellist;

labellist = labellist_start;
while (labellist < copyptr)			    /* stackchange: no modification */
	{
	if (unify(labellist, label) == Success)
		return((INSTRPTR)(labellist[1]));
	labellist += 2;
	}
return((INSTRPTR) 0);
}


static INSTRPTR find_forward_label(MWPTR label, INSTRPTR instrptr)
{
MWPTR labellist;

labellist = copyptr;                    /* stackchange: no modification */
while (labellist > labellist_start)
	{
	labellist -= 2;
	if ((unify(labellist, label) == Success) &&
		((INSTRPTR)(labellist[1]) > instrptr))
		return((INSTRPTR)(labellist[1]));
	}
return((INSTRPTR) 0);
}


static INSTRPTR find_backward_label(MWPTR label, INSTRPTR instrptr)
{
MWPTR labellist;

labellist = labellist_start;
while (labellist < copyptr)				  /* stackchange: no modification */
	{
	if ((unify(labellist, label) == Success) &&
		((INSTRPTR)(labellist[1]) <= instrptr))
		return((INSTRPTR)(labellist[1]));
	labellist += 2;
	}
return((INSTRPTR) 0);
}


static int insert_label(MW m, INSTRPTR labelptr)
{
MWPTR   args;
MW      arity;
int     i;

ValDeref(m);
if (Struct(m))
	{
	args = XtrStruct(m);
	m = FunctIndex(XtrFunctor(args));
	arity = XtrArity(args);
	args = FirstArgPtr(args);
	if ((i = find_instr(m)) < 0)
		return(ErrorNumber = BuiltinInstError);
	if (asm_table[i].asm_instr != LabelType)
		return(Fail);
	if (arity != 1)
		return(ErrorNumber = BuiltinInstError);
	m = args[0];
	ValDeref(m);
	if (NotInt(m) && (find_label(args) != 0))
		return(ErrorNumber = BuiltinInstError);
	labellist_start -= 2;
	labellist_start[0] = m;
	labellist_start[1] = (MW) labelptr;
	return(Success);
	}
return(Fail);
}


static MW compute_label(MW m, INSTRPTR instrptr)
{
MWPTR   args, functor;

ValDeref(m);
if (Int(m))
	return(XtrInt(m) << 2);
else if (Struct(m))
	{
	args = XtrStruct(m);
	functor = XtrFunctor(args);
	args = FirstArgPtr(args);
	if (functor == ForwardFunctor)
		return((MW) find_forward_label(args, instrptr));
	else if (functor == BackwardFunctor)
		return((MW) find_backward_label(args, instrptr));
	else if (functor == LabelFunctor)
		return((MW) find_label(args));
	else if (functor == GlobalFunctor)
		return((MW) find_global_label(args[0]));
	}
ErrorNumber = BuiltinInstError;
return(0);
}


#define add_reg_pattern(atom, code, pos) {\
		MW m; int i; \
		m = atom; ValDeref(m); \
		if (NotAtom(m) || ((i = find_reg(m)) < 0)) \
			return(ErrorNumber = BuiltinInstError); \
		code += ((INSTR) reg_table[i].reg_nr) << pos;}


static int assemble(MW m, INSTRPTR codeptr)
{
MWPTR   args;
MWPTR   functor;
MW      arity;
int     i, j;
INSTR   code;

ValDeref(m);
if (Struct(m))
	{
	args = XtrStruct(m);
	m = FunctIndex(XtrFunctor(args));
	arity = XtrArity(args);
	args = FirstArgPtr(args);
	if ((i = find_instr(m)) < 0)
		return(ErrorNumber = BuiltinInstError);
/*
printf("%d\n",i);
*/
	switch (asm_table[i].asm_instr)
		{
		case LabelType:
			return(Fail);
		case LoadType:
			if (arity != 2)
				return(ErrorNumber = BuiltinInstError);
			code = asm_table[i].asm_imm_code;
			add_reg_pattern(args[0], code, 16);
			m = args[1];
			ValDeref(m);
			if (NotStruct(m))
				return(ErrorNumber = BuiltinInstError);
			args = XtrStruct(m);
			functor = XtrFunctor(args);
			args = FirstArgPtr(args);
			if (functor != CscellFunctor)
				return(ErrorNumber = BuiltinInstError);
			add_reg_pattern(args[0], code, 21);
			m = args[1];
			ValDeref(m);
			if (Int(m))
				{
				m = XtrInt(m);
				if ((m > 32767) || (m < -32768L))
					return(ErrorNumber = BuiltinRangeError);
				}
			else if (Struct(m))
				{
				args = XtrStruct(m);
				functor = XtrFunctor(args);
				args = FirstArgPtr(args);
				if ((m = compute_label(args[0], codeptr)) == 0)
					return(ErrorNumber = BuiltinInstError);
				if (functor != LowerFunctor)
					return(ErrorNumber = BuiltinInstError);
				}
			else
				return(ErrorNumber = BuiltinInstError);
			code += (m & 0x0000ffff);
			break;
		case ArithType:
			if (arity != 3)
				return(ErrorNumber = BuiltinInstError);
			code = 0;
			add_reg_pattern(args[1], code, 21);
			m = args[0];
			ValDeref(m);
			if (NotAtom(m) || ((j = find_reg(m)) < 0))
				return(ErrorNumber = BuiltinInstError);
			m = args[2];
			ValDeref(m);
			if (Int(m))
				{
				m = XtrInt(m);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				if ((m > 32767) || (m < -32768L))
					return(ErrorNumber = BuiltinRangeError);
				code += asm_table[i].asm_imm_code + (m & 0x0000ffff);
				}
			else if (Atom(m))
				{
				code += asm_table[i].asm_reg_code;
				code += (((INSTR) reg_table[j].reg_nr) << 11);
				if ((j = find_reg(m)) < 0)
					return(ErrorNumber = BuiltinInstError);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				}
			else if (Struct(m))
				{
				args = XtrStruct(m);
				functor = XtrFunctor(args);
				args = FirstArgPtr(args);
				if ((m = compute_label(args[0], codeptr)) == 0)
					return(ErrorNumber = BuiltinInstError);
				if (functor == UpperFunctor)
					m = m >> 16;
				else if (functor != LowerFunctor)
					return(ErrorNumber = BuiltinInstError);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				code += asm_table[i].asm_imm_code + (m & 0x0000ffff);
				}
			else
				return(ErrorNumber = BuiltinInstError);
			break;
		case LogicalType:
			if (arity != 3)
				return(ErrorNumber = BuiltinInstError);
			code = 0;
			add_reg_pattern(args[1], code, 21);
			m = args[0];
			ValDeref(m);
			if (NotAtom(m) || ((j = find_reg(m)) < 0))
				return(ErrorNumber = BuiltinInstError);
			m = args[2];
			ValDeref(m);
			if (Int(m))
				{
				m = XtrInt(m);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				if ((m > 65535) || (m < 0))
					return(ErrorNumber = BuiltinRangeError);
				code += asm_table[i].asm_imm_code + (m & 0x0000ffff);
				}
			else if (Atom(m))
				{
				code += asm_table[i].asm_reg_code;
				code += (((INSTR) reg_table[j].reg_nr) << 11);
				if ((j = find_reg(m)) < 0)
					return(ErrorNumber = BuiltinInstError);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				}
			else if (Struct(m))
				{
				args = XtrStruct(m);
				functor = XtrFunctor(args);
				args = FirstArgPtr(args);
				if ((m = compute_label(args[0], codeptr)) == 0)
					return(ErrorNumber = BuiltinInstError);
				if (functor == UpperFunctor)
					m = m >> 16;
				else if (functor != LowerFunctor)
					return(ErrorNumber = BuiltinInstError);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				code += asm_table[i].asm_imm_code + (m & 0x0000ffff);
				}
			else
				return(ErrorNumber = BuiltinInstError);
			break;
		case SubType:
			if (arity != 3)
				return(ErrorNumber = BuiltinInstError);
			code = 0;
			add_reg_pattern(args[1], code, 21);
			m = args[0];
			ValDeref(m);
			if (NotAtom(m) || ((j = find_reg(m)) < 0))
				return(ErrorNumber = BuiltinInstError);
			m = args[2];
			ValDeref(m);
			if (Int(m))
				{
				m = - XtrInt(m);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				if ((m > 32767) || (m < -32768L))
					return(ErrorNumber = BuiltinRangeError);
				code += asm_table[i].asm_imm_code + (m & 0x0000ffff);
				}
			else if (Atom(m))
				{
				code += asm_table[i].asm_reg_code;
				code += (((INSTR) reg_table[j].reg_nr) << 11);
				if ((j = find_reg(m)) < 0)
					return(ErrorNumber = BuiltinInstError);
				code += (((INSTR) reg_table[j].reg_nr) << 16);
				}
			else
				return(ErrorNumber = BuiltinInstError);
			break;
		case NorType:
			code = asm_table[i].asm_reg_code;
			add_reg_pattern(args[0], code, 11);
			add_reg_pattern(args[1], code, 21);
			add_reg_pattern(args[2], code, 16);
			break;
		case LoadImmType:
			if (arity != 2)
				return(ErrorNumber = BuiltinInstError);
			code = asm_table[i].asm_imm_code;
			add_reg_pattern(args[0], code, 16);
			m = args[1];
			ValDeref(m);
			if (Int(m))
				{
				m = XtrInt(m);
				if ((m > 32767) || (m < -32768L))
					return(ErrorNumber = BuiltinRangeError);
				code += (m & 0x0000ffff);
				}
			else if (Struct(m))
				{
				args = XtrStruct(m);
				functor = XtrFunctor(args);
				args = FirstArgPtr(args);
				if ((m = compute_label(args[0], codeptr)) == 0)
					return(ErrorNumber = BuiltinInstError);
				if (functor == UpperFunctor)
					m = m >> 16;
				else if (functor == UpperPlusFunctor)
					m = (m + ((m & 0x00008000) << 1)) >> 16;
				else if (functor != LowerFunctor)
					return(ErrorNumber = BuiltinInstError);
				code += (m & 0x0000ffff);
				}
			else
				return(ErrorNumber = BuiltinInstError);
			break;
		case ShiftType:
			if (arity != 3)
				return(ErrorNumber = BuiltinInstError);
			code = 0;
			add_reg_pattern(args[0], code, 11);
			add_reg_pattern(args[1], code, 16);
			m = args[2];
			ValDeref(m);
			if (Int(m))
				{
				m = XtrInt(m);
				code += asm_table[i].asm_imm_code + ((m & 31) << 6);
				}
			else if (Atom(m))
				{
				code += asm_table[i].asm_reg_code;
				if ((j = find_reg(m)) < 0)
					return(ErrorNumber = BuiltinInstError);
				code += (((INSTR) reg_table[j].reg_nr) << 21);
				}
			else
				return(ErrorNumber = BuiltinInstError);
			break;
		case MulDivType:
			code = asm_table[i].asm_reg_code;
			add_reg_pattern(args[0], code, 21);
			add_reg_pattern(args[1], code, 16);
			break;
		case MovHiLoType:
			code = asm_table[i].asm_reg_code;
			add_reg_pattern(args[0], code, 11);
			break;
		case BranchType:
			if (arity != 3)
				return(ErrorNumber = BuiltinInstError);
			code = asm_table[i].asm_imm_code;
			add_reg_pattern(args[0], code, 21);
			add_reg_pattern(args[1], code, 16);
			if ((m = compute_label(args[2], codeptr)) == 0)
				return(ErrorNumber = BuiltinInstError);
			m = (INSTRPTR) m - 1 - codeptr;
			if ((m > 32767) || (m < -32768L))
				return(ErrorNumber = BuiltinRangeError);
			code += (m & 0x0000ffff);
			break;
		case BranchZeroType:
			if (arity != 2)
				return(ErrorNumber = BuiltinInstError);
			code = asm_table[i].asm_imm_code;
			add_reg_pattern(args[0], code, 21);
			if ((m = compute_label(args[1], codeptr)) == 0)
				return(ErrorNumber = BuiltinInstError);
			m = (INSTRPTR) m - 1 - codeptr;
			if ((m > 32767) || (m < -32768L))
				return(ErrorNumber = BuiltinRangeError);
			code += (m & 0x0000ffff);
			break;
		case JumpType:
			if (arity != 1)
				return(ErrorNumber = BuiltinInstError);
			if ((m = compute_label(args[0], codeptr)) != 0)
				{
				code = asm_table[i].asm_imm_code + ((m >> 2) & 0x03ffffffL);
				}
			else
				{
				code = asm_table[i].asm_reg_code;
				add_reg_pattern(args[0], code, 21);
				}
			break;
		case JumpLinkType:
			if (arity == 1)
				{
				if ((m = compute_label(args[0], codeptr)) != 0)
					{
					code = asm_table[i].asm_imm_code + ((m >> 2) & 0x03ffffffL);
					}
				else
					{
					code = asm_table[i].asm_reg_code + 0x0000f800L;
					add_reg_pattern(args[0], code, 21);
					}
				}
			else if (arity == 2)
				{
				code = asm_table[i].asm_reg_code;
				add_reg_pattern(args[0], code, 11);
				add_reg_pattern(args[1], code, 21);
				}
			else
				return(ErrorNumber = BuiltinInstError);
			break;
		case MoveType:
			code = asm_table[i].asm_reg_code;
			if (arity != 2)
				return(ErrorNumber = BuiltinInstError);
			add_reg_pattern(args[0], code, 11);
			add_reg_pattern(args[1], code, 21);
			break;
		default:
			return(ErrorNumber = BuiltinInstError);
		}
	*codeptr = (MW) code;
	return(Success);
	}
else if (Atom(m))
	{
	m = XtrAtomIndex(m);
	if ((i = find_instr(m)) < 0)
		return(ErrorNumber = BuiltinInstError);
	if (asm_table[i].asm_instr == NopType)
		{
		code = asm_table[i].asm_reg_code;
		*codeptr = (MW) code;
		return(Success);
		}
	}
return(ErrorNumber = BuiltinInstError);
}


/* calls the assembler */
int b_assemble(MW arg[])
{
MWPTR       mptr;
MW	  m, list;
INSTRPTR    instrptr, startptr;
int	 res;

m = *(mptr = arg + 1);
StructDeref(m, mptr);
if (Undef(m))
	{
	Trail(mptr);
	*mptr = MakeInt(((MW) heapptr) >> 2);
	startptr = (INSTRPTR) heapptr;
	}
else if (Int(m))
	{
	startptr = (INSTRPTR) (XtrInt(m) << 2);
	if (((MWPTR) startptr < HeapStart) || ((MWPTR) startptr >= heapptr))
		return(ErrorNumber = BuiltinRangeError);
	}
else
	return(ErrorNumber = BuiltinInstError);
labellist_start = copyptr;				  /* stackchange: no modification */
m = arg[0];
ValDeref(m);
list = m;
instrptr = startptr;
while (List(list))
	{
	mptr = XtrList(list);
	if ((res = insert_label(*mptr++, instrptr)) == Fail)
		instrptr++;
	else if (res != Success)
	list = *mptr;
	ValDeref(list);
	}
if (NotNil(list))
	return(ErrorNumber = BuiltinInstError);
InstrStackCheck;
instrptr = startptr;
while (List(m))
	{
	mptr = XtrList(m);
	if ((res = assemble(*mptr++, instrptr)) == Success)
		instrptr++;
	else if (res != Fail)
	m = *mptr;
	ValDeref(m);
	}
if ((MWPTR) instrptr > heapptr)
	heapptr = ((MWPTR)((((MW)instrptr)+(int)SIZEMW-sizeof(INSTR))&((MW)-(int)SIZEMW)));

#ifdef mips
	cacheflush(startptr, (char *) instrptr - (char *) startptr, BCACHE);
#endif

return(Success);
}


/* extracts the atom arg[0] */
int b_xtr_atom(MW arg[])
{
MWPTR       mptr;
MW	  m, atom;

atom = arg[0];
ValDeref(atom);
if (NotAtom(atom))
	return(ErrorNumber = BuiltinInstError);
m = *(mptr = arg + 1);
StructDeref(m, mptr);
if (NotUndef(m))
	return(ErrorNumber = BuiltinInstError);
Trail(mptr);
m = atom >> 16;
atom = (atom << 16) >> 16;
if (((m != 0) || (atom < 0)) && ((m != -1) || (atom >= 0)))
	{
	if (atom < 0)
		m++;
	CreateList(mptr);
	*mptr++ = MakeInt(m);
	}
*mptr = MakeInt(atom);
return(Success);
}


/* extracts the functor 'functor' */
int b_xtr_functor(MW arg[])
{
MWPTR       mptr;
MW          m, functor;

functor = arg[0];
ValDeref(functor);
if (NotStruct(functor))
	return(ErrorNumber = BuiltinInstError);
functor = (MW) XtrFunctor(XtrStruct(functor));
m = *(mptr = arg + 1);
StructDeref(m, mptr);
if (NotUndef(m))
	return(ErrorNumber = BuiltinInstError);
Trail(mptr);
m = functor >> 16;
functor = (functor << 16) >> 16;
if (((m != 0) || (functor < 0)) && ((m != -1) || (functor >= 0)))
	{
	if (functor < 0)
		m++;
	CreateList(mptr);
	*mptr++ = MakeInt(m);
	}
*mptr = MakeInt(functor);
return(Success);
}


/* jmp */
int b_jump(MW arg[])
{
MW      m;
FUNC    fptr;

m = arg[0];
ValDeref(m);
if (NotInt(m))
	return(ErrorNumber = BuiltinInstError);
fptr = (FUNC) (XtrInt(m) << 2);
m = arg[1];
if (Undef(m))
	m = MakeRef(arg+1);
return(fptr(m));
}


int b_dump(MW arg[])
{
MWPTR   mptr;
MW      m, l;

m = arg[0];
ValDeref(m);
if (NotInt(m))
	return(ErrorNumber = BuiltinInstError);
mptr = (MWPTR) (XtrInt(m) << 2);
l = arg[1];
ValDeref(l);
if (NotInt(l))
	return(ErrorNumber = BuiltinInstError);
l = XtrInt(l);
while (--l >= 0)
	{
	m = *mptr;
	printf("%8lX(%08lX): %08lX   ", (((long) mptr) >> 2), (long) mptr, (long) m);
	mptr++;
	printf(" %1d%1d%1d%1d%1d%1d(%02o)",
		(int) ((m >> 31) & 1), (int) ((m >> 30) & 1),
		(int) ((m >> 29) & 1), (int) ((m >> 28) & 1),
		(int) ((m >> 27) & 1), (int) ((m >> 26) & 1),
		(int) ((m >> 26) & 63));
	printf(" %1d%1d%1d%1d%1d(%2s)",
		(int) ((m >> 25) & 1), (int) ((m >> 24) & 1),
		(int) ((m >> 23) & 1), (int) ((m >> 22) & 1),
		(int) ((m >> 21) & 1), reg_names[(int) ((m >> 21) & 31)]);
	printf(" %1d%1d%1d%1d%1d(%2s)",
		(int) ((m >> 20) & 1), (int) ((m >> 19) & 1),
		(int) ((m >> 18) & 1), (int) ((m >> 17) & 1),
		(int) ((m >> 16) & 1), reg_names[(int) ((m >> 16) & 31)]);
	printf(" %1d%1d%1d%1d%1d(%2s)",
		(int) ((m >> 15) & 1), (int) ((m >> 14) & 1),
		(int) ((m >> 13) & 1), (int) ((m >> 12) & 1),
		(int) ((m >> 11) & 1), reg_names[(int) ((m >> 11) & 31)]);
	printf(" %1d%1d%1d%1d%1d",
		(int) ((m >> 10) & 1), (int) ((m >> 9) & 1),
		(int) ((m >> 8) & 1), (int) ((m >> 7) & 1),
		(int) ((m >> 6) & 1));
	printf(" %1d%1d%1d%1d%1d%1d(%02o)\n",
		(int) ((m >> 5) & 1), (int) ((m >> 4) & 1),
		(int) ((m >> 3) & 1), (int) ((m >> 2) & 1),
		(int) ((m >> 1) & 1), (int) (m & 1), (int) (m & 63));
	}
return(Success);
}


/* extracts the heapptr */
int b_xtr_heapptr(MW arg[])
{
MWPTR   mptr;
MW      m;

m = *(mptr = arg);
StructDeref(m, mptr);
if (Undef(m))
	{
	Trail(mptr);
	*mptr = MakeInt(((MW) heapptr) >> 2);
	return(Success);
	}
else 
	return(ErrorNumber = BuiltinInstError);
}

